import os
import urllib.request
from tqdm import tqdm
from colorama import init, Fore, Back, Style
import re

init()

# for prod only!
import warnings
warnings.filterwarnings("ignore")

from .API_helper import *
from .config import Config


class OpenCity:
    """
    docstring for OpenCity
    """

    # current_list = pd.read_csv(CURRENT_PACKAGE_LIST_FILE)
    def __init__(self, cf=None, interactive = True):
        self.formats = ["csv", "json", "zip", "xls", "txt", "geojson", "kml", "xlsx"]
        self.names_file = "" 
        self.current_packages_file = ""
        self._interactive=interactive
        if not cf:
            cf = Config()
        self.cf = cf
        self.dsuf = DataSetUrlFetcher(cf,interactive=self._interactive)
        self.id_helper = IdHelper(self.dsuf)
        self.show_data_helper = ShowDataHelper(self.dsuf.current_list)


    def get_data(self, data=[], tag=False, meta=False):
        """
        general get data function to be called by user

        PARAMETERS:
        -----------
        data: list of Strings
            containing names or tags (string)
        tag: Boolean
            default: False
            set to True if data list contains tags
        meta : Boolean
            default: False

        RETURNS:
        -----------
        DataFrame|dict: single df or dict containing df
        """
        # if not hasattr(self, 'id_list'):
        self.id_list, spelling = self.id_helper.create_id_list(data, tag)

        result_dict = {}
        #final_flag = True
        
        if spelling:

            if meta:
                print("Loading meta data")
                length = len(self.id_list)
                df_meta = pd.DataFrame(columns=['id', 'url', 'format', 'name', 'created', 'last_modified', 'description'])
                for i in range(length):#tqdm(range(length), total=length, desc=f"[#] "):
                    for id, url, format, name, created, last_modified, description in FetchHelper.fetch_dataset_meta(self.id_list[i]):
                        df_meta = df_meta.append({'id': id, 'url': url, 'format': format, 'name': name, 'created': created, 'last_modified': last_modified, 'description': description}, ignore_index=True)
                result_dict["meta"] = df_meta
                tqdm.write(f"{Fore.GREEN}[+]{Style.RESET_ALL} Successfully loaded meta data of {df_meta.shape[0]} data sets")

                    
            else:
                print("Loading data")
                length = len(self.id_list)
                for i in tqdm(range(length), total=length, desc=f"[#] "):
                    for url, format, name in FetchHelper.fetch_dataset_urls(self.id_list[i]):
                        ending = FetchHelper.get_url_ending(url) # works also with Kn Gis Hub?
                        if ending in self.formats:
                            instance = FetchHelper.get_instance(ending)()
                            df, flag = instance.load_data(url)
                        #print(flag)

                            if flag:
                            # print("Successfully loaded data set: " + key)

                                key = (name + "_" + format).replace(" - ", "_").replace(" ", "_").lower()
                                result_dict[key] = df

                                tqdm.write(f"{Fore.GREEN}[+]{Style.RESET_ALL} Successfully loaded data set:\t {key}")


                            else:
                            # print("Data set was omitted: " + key)
                                output_name = name + "_" + format
                                tqdm.write(f"{Fore.RED}[x]{Style.RESET_ALL} Data set was omitted:\t\t {output_name}")
                        else:
     
                            tqdm.write(f"{Fore.YELLOW}[-]{Style.RESET_ALL} External Link:\t\t\t {name}\n\t\t\t\t\t Please visit {url}")

        if len(result_dict) == 1:
            key = list(result_dict.keys())[0]
            result = result_dict[key]
        else:
            result = result_dict

        return result


    def save_data(self, data, tag=False, folder="", file_ret=False):
        """
        function to save the indicated data (or data fitting the indicated tags)
        to your local disk

        PARAMETERS:
        -----------
        data: list of Strings
            list containing names of the datasets you want to store
            or tags for which you want to save the respective datasets
        tag: Boolean
            default: False
            set to True if data list contains tags
        folder: String
            default: empty
            if you wanted to save the data to a different folder than the one from which you are executing the python file,
            you could indicate the respective folder here (use either forward slashes '/' or double backward slashes '\\')
        file_ret: Boolean
            default: False
            for testing purposes

        RETURNS:
        -----------
        void
        """
        # if not self.id_list:            
        self.id_list, spelling = self.id_helper.create_id_list(data, tag)

        url_list = []
        key_list = []
        file_return = []
        if spelling: 
            for i in range(len(self.id_list)):
                for url, format, name in FetchHelper.fetch_dataset_urls(self.id_list[i]):
                    ending = FetchHelper.get_url_ending(url) 
                    if ending in self.formats:
                        url_list.append(url)
                        key_list.append(re.sub("[*:/<>?\|]", "-", name) + "." + ending) # removing special characters not appropriate for file names

            for url, key in zip(url_list, key_list):
                file_name = ""
                # if a specific folder was given:
                if len(folder)>0: 
                    #if the indicated directory exists
                    if os.path.isdir(folder): 
                        file_name = os.path.join(folder, key)
                    #if the indicated directory doesn't exist already
                    elif self._interactive: 
                        print(
                            f"{Fore.RED}There is no folder with the name {folder}.{Style.RESET_ALL}"
                        )
                        inp = input(
                            "Do you wish to proceed (and let it be created)? [y/N]\n> "
                        )
                        if inp == "N":
                            print(f"{Fore.RED}> EXITING{Style.RESET_ALL}")
                            #sys.exit(0) #TODO is it necessary to throw the user out of python?
                            break
                        elif inp == "y":
                            print(f"{Fore.RED}> CREATING DIRECTORY{Style.RESET_ALL}")
                            os.mkdir(path = folder)
                            file_name = os.path.join(folder, key)                    
                    else:
                        os.mkdir(path = folder)
                        file_name = os.path.join(folder, key)                    
                # if no folder was given: save to current working directory
                else:
                    file_name = os.path.join(os.getcwd(), key)
                # if the url leads to a file: 
                if url[-5:] != "=json":
                    urllib.request.urlretrieve(url, file_name) 
                    print("Finished saving requested data to " + file_name)
                    file_return.append(file_name)
                # if the url is the result of a get query for a geojson: 
                else:                
                    geodf = shpFetcher()
                    geodf.parse_geo(url).to_file(file_name, driver="GeoJSON")
                    print("Finished saving requested data to " + file_name)
                    file_return.append(file_name)
        if file_ret:
            return file_return
        return ""


    def show_data(self, data = [], tag = False, overview = False, meta = False, terminal = False): 
        """
        function to get an overview of the data sets available

        PARAMETERS:
        -----------
        overview: Boolean
            default: False
            set to True if you wanted to get a short overview (title, short name, tags)
            of the datasets in your console
        meta: Boolean
            default: False
            set to True if you wanted more detailed information on the datasets
            depending on paremeter 'terminal', whether you get the output in your console or as a popup
        data: list of Strings
            list containing names of the datasets you want to store
            or tags for which you want to save the respective datasets
        tag: Boolean
            default: False
            set to True if data list contains tags
        terminal: Boolean
            default: False
            set to True if you want to print the meta data in your console instead of a popup

        RETURNS:
        -----------
        void
        """

        if overview == True and meta == True or overview == False and meta == False and len(data) > 0 or isinstance(tag, list): 
            tqdm.write(f"{Fore.RED}You did not use the function correctly.")
            tqdm.write(f"{Fore.RED}Use either the parameter 'overview' OR the parameter 'meta'.")
            tqdm.write(f"{Fore.RED}Use the parameter 'data' in addition if you only want to get an overview or the metadata for specific datasets or tags.")
            tqdm.write(f"{Fore.RED}Set the parameter 'tag' to True if you indicated tags instead of single datasets under the 'data' parameter.{Style.RESET_ALL} ")

        # if no input is given: 
        elif overview == False and meta == False and len(data) == 0: 
            self.show_data_helper.summary()

        # if no dataset or tag is given: 
        elif len(data) == 0: 
            if overview == True: 
                self.show_data_helper.short(self.dsuf.current_list)
            if meta == True:
                self.show_data_helper.summary()
                if terminal: 
                    print("\nIn the following you will see detailed information on all the datasets:\n")
                    self.show_data_helper.long(self.dsuf.current_list)
                else:
                    if TK:             
                        print("\nIn the following popup you will see detailed information on all the datasets:")
                        self.show_data_helper.meta(self.dsuf.current_list)
                    else:
                        print(f"{Fore.RED}There is an error with your Tkinter installation, use terminal=True to show the information anyway.{Style.RESET_ALL}")

        # if a dataset or a tag is given: show only the indicated datasets
        elif len(data) > 0: 
            #check spelling first:
            self.id_list, spelling = self.id_helper.create_id_list(data, tag)
            if spelling: 
                tag_df = pd.DataFrame(columns = ['title', 'name', 'tags'])
                if tag: #if a tag is given
                    for element in data: 
                        tag_df = pd.concat([tag_df, self.dsuf.current_list[self.dsuf.current_list.tags.str.contains(element)]], axis = 0) # create df containing only the data sets with that tag
                else: #if a dataset is given
                    for element in data: 
                        tag_df = pd.concat([tag_df, self.dsuf.current_list[self.dsuf.current_list.name.str.contains(element)]], axis = 0)
                tag_df = tag_df.drop_duplicates(subset = "title")
                if overview == True: 
                    self.show_data_helper.short(tag_df)
                if meta == True:
                    if terminal: 
                        print("In the following you will see detailed information on {}:".format(data))
                        self.show_data_helper.long(tag_df)
                    else: 
                        if TK:                            
                            print("In the following popup you will see detailed information on {}:".format(data))
                            self.show_data_helper.meta(tag_df)
                        else:
                            print(f"{Fore.RED}There is an error with your Tkinter installation, use terminal=True to show the information anyway.{Style.RESET_ALL}")
        
        else: 
            tqdm.write(f"{Fore.RED}You did not use the function correctly.")
            tqdm.write(f"{Fore.RED}Use either the parameter 'overview' OR the parameter 'meta'.")
            tqdm.write(f"{Fore.RED}Use the parameter 'data' in addition if you only want to get an overview or the metadata for specific datasets or tags.")
            tqdm.write(f"{Fore.RED}Set the parameter 'tag' to True if you indicated tags instead of single datasets under the 'data' parameter.{Style.RESET_ALL} ")

# TODO: check if really want to download (disable with param)



import os
import sys
import time
import math
import requests
import pandas as pd
from tqdm import tqdm
from bs4 import BeautifulSoup
from colorama import init, Fore, Back, Style

init()  # colorama

# from .config import Config as cf


class DataSetUrlFetcher:
    """
	Handles the fetching and storing of all available datasets
	"""
    def __init__(self, cf, interactive = True):
        self.cf = cf
        self._interactive = interactive
        self.current_list = self.read_curr_packages()

    def read_curr_packages(self):
        try:
            data_frame = pd.read_csv(self.cf.CURRENT_PACKAGE_LIST_FILE)
        except Exception as e:
            if self._interactive:
                print(
                    f"{Fore.RED}There is no file with name {self.cf.CURRENT_PACKAGE_LIST_FILENAME} in the directory: {self.cf.PKG_FOLDER}{Style.RESET_ALL}"
                )
                inp = input(
                    "The file is needed, do you wish to proceed (and let it be created)? [y/N]\n> "
                )
                if inp.lower() == "n":
                    print(f"{Fore.RED}> EXITING")
                    sys.exit(0)
                    return 0
            resp = self.fetch()
            if isinstance(resp, int):
                print(f"Error: status_code = {resp}")
                return False

            data_frame = self._parse_data(resp)
            self._store(data_frame)
            
        return data_frame

    def fetch(self):
        """
		basic fetch method for the self.cf.CURRENT_PACKAGE_LIST_URL

		PARAMETERS:
		-----------
		None

		RETURNS:
		-----------
		Json: current packages (success)
		Int: Status code (error)
		"""
        response = requests.get(self.cf.CURRENT_PACKAGE_LIST_URL)  # , header =
        if response.status_code == 200:
            return response.json()
        return response.status_code
        
    def _get_names(self):
        try:
            # os.path.join(self.cf.CWD, 'names.csv')
            names = pd.read_csv(self.cf.GH_NAMES_FILE_URL, sep=";")
        except Exception as e:
            print(
                f"{Fore.RED}An error occured while trying to read the names to id file: {Style.RESET_ALL}\n {e}"
            )
            # names.to_csv(self.cf.NAMES_FILENAME, index=False)
            raise e
        return names
            
            
            
    def _store(self, data_frame: pd.DataFrame) -> bool:
        """
		writes dataframe to file

		PARAMETERS:
		-----------
		data_frame: DataFrame
			the respective DataFrame to store

		RETURNS:
		-----------
		sucess: Boolean
			indicates wether the storing was successfull
		"""
        if not isinstance(data_frame, pd.DataFrame):
            print(f"Expected DataFrame, got {type(data_frame)}")
            return False

        try:
            
            #name_list = self._get_names()

            # name_list = pd.read_csv(names_file, sep=';')
            #merged_list = pd.merge(data_frame, name_list, how='left', on='id')
            #if merged_list['name'].isnull().values.any():
            #    idx = merged_list.index[merged_list['name'].isnull()].tolist()
            #    print(idx)
            data_frame.to_csv(self.cf.CURRENT_PACKAGE_LIST_FILE,
                               encoding='utf-8',
                               index=False)
            return True
        except Exception as writing_file_error:
            print(writing_file_error)
            return False

    def _parse_data(self, data):
        """
		parse data from json into DataFrame

		PARAMETERS:
		-----------
		data: string (json)
			json string fetched for a resource

		RETURNS:
		-----------
			DataFrame with all info
		"""
        if not "success" in data:
            return False

        results = data["result"][0]
        
        out = list()
        for item in tqdm(results):
            tags = []
            try:
                for tag_item in item["tags"]:
                    tags.append(tag_item["name"])
                out.append({
                    "id":
                    item["id"],
                    "title":
                    item["title"],
                    "source":
                    item["url"],
                    "url":
                    "https://offenedaten-konstanz.de/api/3/action/package_show?id="
                    + item["id"],
                    "created":
                    item["metadata_created"],
                    "modified":
                    item["metadata_modified"],
                    "notes":
                    item["notes"],
                    #BeautifulSoup(item["notes"], "lxml").text,
                    "tags":
                    tags
                })
            except:
                item_name = item["name"]
                print(item["name"] + " item has not all information needed, hence omitted."
                      )  # concerns unvollständige data sets on OpenData website
        
        data_frame = pd.DataFrame.from_dict(out)
        name_list = self._get_names()

            # name_list = pd.read_csv(names_file, sep=';')
        merged_list = pd.merge(data_frame, name_list, how='left', on='id')
        if merged_list['name'].isnull().values.any():
            idx = merged_list.index[merged_list['name'].isnull()].tolist()
            for i in idx:
                merged_list['name'][i] = merged_list['title'][i]
            
        return merged_list

    def update(self):
        """
		update method which handles the fetching, parsing
		and storing of the info

		PARAMETERS:
		-----------
		None

		RETURNS:
		-----------
		success: Boolean
			wether the operation was successfull
		"""
        resp = self.fetch()
        if isinstance(resp, int):
            print(f"Error: status_code = {resp}")
            return False

        data_frame = self._parse_data(resp)
        # check if names are missing !!!!
        #print(data_frame)

        #store_status = self._store("st")
        store_status = self._store(data_frame)
        if not store_status:
            print("Error while storing data")
            return False

        return True


# dsuf = DataSetUrlFetcher()
# s = dsuf.update()




import pandas as pd
import xlrd

class xlsFetcher(object):
    """
    xlsFetcher: fetces xls-Data
    """

    def __init__(self):
        self.flag_final = True

    def parse_xls(self, url):
        """
        parses data from url to dataframe

        PARAMETERS:
        -----------
        url: String
            Data ID based link

        RETURNS:
	    -----------
        DataFrame: data for url
        """
        try:
            df = pd.read_excel(url, sheet_name=None)
            return df
        except:
            self.flag_final = False
            return pd.DataFrame()

    def load_data(self, url):
        """
        load data set

        PARAMETERS:
	    -----------
        url: String
			Data ID based link

        RETURNS:
	    -----------
        DataFrame: data for url
        Boolean: flag_final (success)
        """
        return self.parse_xls(url), self.flag_final



import os
import time

from .config import Config as cf
from .fetch_dataset_list import DataSetUrlFetcher

dsuf = DataSetUrlFetcher()

MAX_DAY_DELTA = 14


def check():
    """
    checks if the local version of the current packages is up to date:
    has it been modified/updated in the last 14 days
    if no, ask the user if he wants to update and the proceed

    PARAMETERS:
    -----------
    None

    RETURNS:
    -----------
    void
    """
    if not dsuf.current_list:
        dsuf.update()
        return 1
    modified_time = os.path.getmtime(cf.CURRENT_PACKAGE_LIST_FILE)
    delta = time.time() - modified_time
    day_delta = delta / 60 / 60 / 24
    # print(day_delta)
    if day_delta > MAX_DAY_DELTA:
        prompt = input(
            "Your local list of available data packages is outdated. Do you want to download the latest version? [y/N]"
        ).lower()
        if prompt == "n":
            print("> Not downloading")
            return 0
        print("> Updating packages...")
        dsuf.update()


# check()



import re
from pandas import pandas as pd
from urllib.request import urlopen
from urllib.error import HTTPError

class txtFetcher(object):
    """
    txtFetcher: fetches txt-Data
    """
    def __init__(self):
        self.encoding = ["utf-8", "ISO-8859-1", "latin1"]
        self.encoding_idx = 0 # default value
        self.encoding_final = self.encoding[self.encoding_idx] # default value
        self.flag_final = False
        self.flag_encoding = False
        self.lines_list = []
        self.first_line = False

    def get_encoding(self, file):
        """
        extract encoding from first line if applicable

        PARAMETERS:
        -----------
        file: String
            path to file

        RETURNS:
        -----------
        void
        """
        try:
            first_line = file.readline().decode(self.encoding[self.encoding_idx])
        except:
            print("false encoding")
            self.flag_encoding = True
            if self.encoding_idx < len(self.encoding)-1:
                self.encoding_idx += 1
                self.get_encoding(file)
            else:
                print("no matching encoding found. please review")

        if re.findall("encoding", first_line):
            self.first_line = True
            for i in re.finditer("encoding", first_line):
                text = first_line[i.end():]
                self.encoding_final = str(re.findall('"([^"]*)"', text)[0]).strip('[]')

    def read_line(self, line):
        """
        read single line, cleans it and adds it to the lines_list

        PARAMETERS:
        -----------
        line: String
            line of a file

        RETURNS:
        -----------
        void
        """
        decoded_line = line.decode(self.encoding_final)
        for quoted_part in re.findall(r'\"(.*?)\"', decoded_line):  # replace 'space' in quotes with '@'
            decoded_line = decoded_line.replace(quoted_part, quoted_part.replace(" ", "@"))
        split_lines = decoded_line.split(' ')
        for i in range(len(split_lines)):
            for quoted_part in re.findall(r'\"(.*?)\"',
                                          split_lines[i]):  # replace '@' in quotes with 'space'
                split_lines[i] = split_lines[i].replace(quoted_part, quoted_part.replace("@", " "))
        self.lines_list.append(split_lines)

    def get_data(self, url):
        """
        gets the data from url

        PARAMETERS:
        -----------
        url: String
            Data ID based link

        RETURNS:
        -----------
        lines_list: list of Strings
        """
        try:
            file = urlopen(url)
        except HTTPError:
            print("could not open url")
            self.flag_final = True

        if self.flag_final == False:
            try:
                self.get_encoding(file)
                self.encoding_idx = 0 # reset to default value
                for line in file:
                    try:
                        self.read_line(line)
                    except Exception as read_error:
                        print("false encoding")
                        self.flag_encoding = True
                        if self.encoding_idx < len(self.encoding) - 1:
                            self.encoding_idx += 1
                            self.encoding_final = self.encoding[self.encoding_idx]
                            self.read_line(line)
                        else:
                            print("no matching encoding found. please review")
            except:
                self.flag_final = True

            return self.lines_list

    def convert_df(self, data):
        """
        converts the data to a Dataframe

        INPUT:
        -----------
        data: String
            txt. data

        OUTPUT:
        -----------
        DataFrame: data as DataFrame
        """
        df = pd.DataFrame(data)
        row, col = df.shape
        drop_cols = []
        for i in range(col):
            if not df.iloc[:,i].str.contains("=").any(): # if col does not contain data
                drop_cols.append(i)
        if len(drop_cols) != 0:
            df = df.drop(drop_cols, axis=1)  # drop respective cols

        row, col = df.shape
        col_name_ref = ''  # default value
        col_names = []

        for j in range(col):  # get col names and extract values
            for i in range(row):
                if i == 0:
                    col_name_ref = df.iloc[i, j].split("=")[0] # get col name
                split_element = df.iloc[i, j].split("=")
                col_name = split_element[0]

                if j == (col - 1): # handling last col
                    value = split_element[1].split("/>")[0][1:-1]
                else:
                    value = split_element[1][1:-1]

                if col_name_ref != col_name: # case of different col names
                    print("column names do not match. please review error")
                    break

                if value == "": # handling missing data
                    df.iloc[i, j] = "NaN"
                else:
                    df.iloc[i, j] = value
                print(f"row {i}")

            col_names.append(col_name_ref)
        df.columns = col_names
        return df

    def load_data(self, url):
        """
        function to load the data, called by external classes

        PARAMETERS:
        -----------
        url: String
            Data ID based link

        RETURNS:
        -----------
        DataFrame: data for url
        Boolean: flag_final (success)
        """
        data = self.get_data(url)
        return self.convert_df(data), self.flag_final




# import os
from pandas import pandas as pd
from .basic_fetcher import BasicFetcher

class csvFetcher(BasicFetcher):
	"""
	csvFetcher
	inherits basic fetcher
	"""
	def __init__(self):
		self.flag_final = False
		self.fallback_seperator = ";"

	def verify_df(self, df1, url, encoding, sep):
		"""
		checkf if csv has been correctly read into dataframe
		-> avoid issues

		PARA METERS:
	    -----------
		df1: Dataframe
			current df1
		url: String
			data ID based link
		encoding: String
			encoding for csv file
		sep: String
			csv file seperator

		RETURNS:
	    -----------
		Boolean: if dataframe is created correctly
		"""
		try:
			anti_sep = "," if sep == ";" else ";"
			df2 = pd.read_table(url, sep=anti_sep,
									encoding=encoding,
									engine="python",
									keep_default_na=False,
									skip_blank_lines=True
									)
			if len(list(df2.columns)) > len(list(df1.columns)):
				#print("+++ False")
				return False
		except:
			pass
			#print("*** TRUE")
		return True


	def parse_csv(self, url, encoding="utf-8", det_flag=False, decode_flag=False, sep=";"):
		"""
		parses data from url to dataframe

		PARAMETERS:
	    -----------
		url: String
			Data ID based link
		encoding: String
			default: 'utf-8'
			encoding for csv file
		det_flag: Boolean
			default: False
			control max number of repetitions
		decode_flag: Boolean
			default: False
			track if another encoding has been tried
		sep: String
			default: ';'
			csv file seperator

		RETURNS:
	    -----------
		DataFrame: data for url
        Boolean: flag_final (success)
		"""
		try:
			df = pd.read_csv(url, sep=sep,
									encoding=encoding,
									engine="python",
									keep_default_na=False,
									skip_blank_lines=True
									# error_bad_lines=True,
									# decimal=",",
									# quotechar='"',
									)
			if self.verify_df(df, url, encoding, sep):
				self.flag_final = True
				return df
			raise Exception("SeperatorException: expected after")
		except Exception as read_error:
			if not decode_flag and isinstance(read_error, UnicodeDecodeError):
				# iso-8859-1
				return self.parse_csv(url, encoding='latin1', decode_flag=True, sep=sep)
			if not det_flag and "expected" in str(read_error).lower():
				return self.parse_csv(url, encoding=encoding, det_flag=True, sep=",")
			return None, False


	def load_data(self, url):
		"""
        function to load the data

        PARAMETERS:
	    -----------
        url: String
			Data ID based link

        RETURNS:
	    -----------
        DataFrame: data for url
        Boolean: flag_final (success)
        """
		return self.parse_csv(url), self.flag_final




import requests

class BasicFetcher:
    """
    BasicFetcher:
    works as a parent class for
    all file-specific fetcher
    """

    def __init__(self):
        """
        - sets header (dict)
        - sets package base url
        """
        self.PACKAGE_BASE_URL = "https://offenedaten-konstanz.de/api/3/action/package_show?id="
        self.headers = {
            "User-Agent": "PythonOpenDataPackage/1.0"
        }

    def fetch_resource_urls(self, p_id):
        """
        Generator for requests to API

        PARAMETERS:
        -----------
        p_id: String
            package id

        RETURNS:
        -----------
        yield: url, name, format
        if staus code not 200
            return status code
        """
        resp = requests.get(self.PACKAGE_BASE_URL + p_id, headers=self.headers)
        if resp.status_code == 200:
            resources = resp.json()["result"][0]["resources"]
            for resource in resources:
                yield resource["url"], resource["name"], resource["format"]
        return resp.status_code

    def get_file_ending(self, url):
        """
        extract file ending from an url

        PARAMETERS:
        -----------
        url: String
            any url

        RETURNS:
        -----------
        file ending for any url/file
        """
        return url.split(".")[::-1][0].lower()



import json
import requests
from pandas import pandas as pd
import geopandas as gpd  # install

class shpFetcher:
    """
    SHPFetcher: fetches GeoData
    """

    def __init__(self):
        self.flag_final = True

    def parse_geo(self, url):
        """
        parses data from url to dataframe

        PARAMETERS:
        -----------
        url: String
            Data ID based link

        RETURNS:
        -----------
        DataFrame: data for url
        """
        try:
            df = gpd.read_file(url)
            #print("was fine")
            return df
        except:
            self.flag_final = False
            return gpd.GeoDataFrame()

    def load_data(self, url):
        """
        load data set for url

        PARAMETERS:
        -----------
        url: String
            Data ID based link

        RETURNS:
        -----------
        DataFrame: data for url
        Boolean: flag_final (success)
        """
        #print(self.flag_final)
        #print(result)
        return self.parse_geo(url), self.flag_final



from opencity import opencity as oc
from opencity import config as conf 


def test_functional_smoke(tmpdir):
    """Smoke Test

    Doesn't  actually test anything but
    catches errors including package setup errors.
    Requires the API to be available.
    """
    cf = conf.Config(PKG_FOLDER=str(tmpdir))
    open_city = oc.OpenCity(cf = cf,interactive=False)
    open_city.get_data(["radverkehr_stadtradeln"])
    



import json
import requests
from pandas import pandas as pd
import geopandas as gpd  # install
import fiona

class kmlFetcher:
    """
    KMLetcher: fetches GeoData (KML)
    """

    def __init__(self):
        self.flag_final = True

    def enable_KML(self):
        """
        enable KML support

        PARAMETERS:
	    -----------
        None

        RETURNS:
	    -----------
        void

        """
        fiona.drvsupport.supported_drivers['kml'] = 'rw'
        fiona.drvsupport.supported_drivers['KML'] = 'rw'

    def parse_geo(self, url):
        """
        parses data from url to dataframe

        PARAMETERS:
        -----------
        url: String
            Data ID based link

        RETURNS:
        -----------
        DataFrame: data for url
        """

        self.enable_KML()

        try:
            df = gpd.read_file(url)
            return df
        except:
            self.flag_final = False
            return gpd.GeoDataFrame()

    def load_data(self, url):
        """
        load data set

        PARAMETERS:
        -----------
        url: String
            Data ID based link

        RETURNS:
	    -----------
        DataFrame: data for url
        Boolean: flag_final (success)
        """
        return self.parse_geo(url), self.flag_final



import os
import sys
import requests
import pandas as pd
import numpy as np
from tabulate import tabulate
import re
try:
    import tkinter as tk
    import tksheet
    TK = True
except Exception as e:
    TK = False

from colorama import init, Fore, Back, Style

init()  # colorama

from .config import Config as cf

from .fetch_dataset_list import DataSetUrlFetcher

from .fetcher.csv_fetcher import csvFetcher
from .fetcher.shp_fetcher import shpFetcher  # applies for some jsons as well
from .fetcher.xls_fetcher import xlsFetcher
from .fetcher.txt_fetcher import txtFetcher
from .fetcher.json_fetcher import jsonFetcher
from .fetcher.kml_fetcher import kmlFetcher

formats_dict = {
    "csv": csvFetcher,
    "txt": txtFetcher,
    "xls": xlsFetcher,
    "xlsx": xlsFetcher,
    "zip": shpFetcher,
    "geojson": shpFetcher,
    "json": jsonFetcher,
    "kml": kmlFetcher
}


# current_list = read_curr_packages()


class FetchHelper:
    """
    helper class containing functions for fetching
    """
    def __init__(self):
        pass

    def get_instance(format):
        """
        gets instance of repective fetcher

        PARAMETERS:
        -----------
        format: String
            file format (e.g. csv, json, ...)

        RETURNS:
        -----------
        Fetcher: respective fetcher
        """
        return formats_dict[format]

    def fetch_dataset_urls(id):
        """
        get urls corresponding to id

        PARAMETERS:
        -----------
        id: String
            package id

        RETURNS:
        -----------
        yield: url, name, format
        if staus code not 200
            return status code
        """
        response = requests.get(cf.PACKAGE_BASE_URL + id)
        if response.status_code == 200:
            resources = response.json()["result"][0]["resources"]
            for resource in resources:
                yield resource["url"], resource["format"], resource["name"]
        return response.status_code
    
    def fetch_dataset_meta(id):
        """
        get meta data corresponding to id

        PARAMETERS:
        -----------
        id: String
            package id

        RETURNS:
        -----------
        yield: id, url, format, name, created, last_modified, description
        if staus code not 200
            return status code
        """
        response = requests.get(cf.PACKAGE_BASE_URL + id)
        if response.status_code == 200:
            resources = response.json()["result"][0]["resources"]
            for resource in resources:
                yield resource["id"], resource["url"], resource["format"], resource["name"], resource["created"], resource["last_modified"], resource["description"]
        return response.status_code

    def get_url_ending(url):
        """
        get ending of url/file

        PARAMETERS:
        -----------
        url: String
            respective url

        RETURNS:
        -----------
        String: file type/url ending
        """
        if url[-5:] == "=json":
            return "geojson"
        else:
            return url.split(".")[::-1][0].lower()

    def verify_url(self, url):
        """
        check if an url belongs to offenedaten-konstanz.de

        PARAMETERS:
        -----------
        id: String
            respective url

        RETURNS:
        -----------
        Boolean: if url has offene-daten as domain
        """
        return "offenedaten-konstanz.de" in url.lower()


class IdHelper:
    """
    helper class for creating id list
    """
    def __init__(self, dsuf):
        self.dsuf = dsuf
        if hasattr(self.dsuf, "current_list"):
            self.current_list = self.dsuf.current_list
        else:
            self.current_list = read_curr_packages()

    def create_id_list(self, data, tag=False):
        """
        helper function to create a list of ids for the datasets
        indicated by the names/tags given

        PARAMETERS:
        -----------
        data: list of Strings
            containing names or tags (string)
        tag: Boolean
            default: False
            set to True if data list contains tags

        RETURNS:
        -----------
        List<String>: list of ids
        """
        id_list = []
        if tag:
            for i in range(len(data)):
                for j in range(len(self.current_list)):
                    if data[i] in self.current_list.loc[j, "tags"]:
                        id_list.append(self.current_list.loc[j, "id"])
        else:
            for i in range(len(data)):
                if data[i] in self.current_list["name"].values:
                    id_element = np.array2string(self.current_list[
                        self.current_list['name'] == data[i]]['id'].values)
                    id_list.append(id_element[2:-2])
        if len(id_list) == 0:
            print("The provided names or tags are incorrect." + 
            "\nPlease check spelling. Note that the names for data sets are written in lower case, whereas tags are written with capital letters. " + 
            "\nAdditionally, make sure to set the parameter 'tag' to True if you specified a tag." + 
            "\nFurthermore, make sure to always provide the names of the data sets or tags as a list of strings.")
            spelling = False
        else:
            spelling = True
            
        return id_list, spelling

class ShowDataHelper: 
    """
    helper class for the show_data() function
    """

    def __init__(self, current_list):
        self.current_list = current_list

    def summary(self):
        tags = []
        for taglist in self.current_list.tags:
            clean = re.findall(r"\'(.*?)\'", taglist) #find everything enclosed by '...'
            for entry in clean: 
                if entry not in tags:
                    tags.append(entry)        
        print('There are in total {} datasets available.\nThese datasets belong to {} different categories.These categories are: {}'.format(len(self.current_list), len(tags), tags)) 
    
    def short(self, df):
        print(tabulate(df[['title', 'name', 'tags']], headers = ['Title', 'Shortname', 'Tags']))

    def long(self, df):
        long = pd.melt(df,id_vars = 'title')
        titles = long.title.unique()
        for element in titles: 
            df_element = long.loc[long['title'] == element]
            print(" \n" + element + " :\n" + tabulate(df_element[['variable', 'value']], headers = ['Variable', 'Value'], showindex=False))
    
    def meta(self, df):
        df = df[['title', 'name', 'id', 'modified', 'source', 'notes', 'tags']]
        df = df.values.tolist()
        headers = ['Title', 'Shortname', 'ID', 'Last edited on', 'Source', 'Notes', 'Tags']
        app = tk.Tk()
        table = tksheet.Sheet(app, height=1000, width = 2000)
        table.grid()
        table.headers(headers)
        table.enable_bindings(("single_select",
                       "row_select",
                       "column_width_resize",
                       "arrowkeys",
                       "right_click_popup_menu",
                       "copy"))
        table.change_theme(theme = "light blue")
        table.set_sheet_data(data = df, reset_highlights = True, reset_col_positions=True, reset_row_positions=True)
        table.set_all_cell_sizes_to_text(redraw = True)
        app.mainloop()




import pandas as pd

class jsonFetcher(object):
    """
    jsonFetcher: fetches json-Data
    """

    def __init__(self):
        self.flag_final = True

    def parse_json(self, url):
        """
        parses data from url to dataframe

        PARAMETERS:
        -----------
        url: String
            Data ID based link

        RETURNS:
        -----------
        DataFrame: data for url
        """
        try:
            df = pd.read_json(url)
            # TO DO : different encodings
            return df
        except:
            self.flag_final = False
            return pd.DataFrame()

    def load_data(self, url):
        """
        function to load the data

        PARAMETERS:
	    -----------
        url: String
			Data ID based link

        RETURNS:
	    -----------
        DataFrame: data for url
        Boolean: flag_final (success)
        """
        return self.parse_json(url), self.flag_final
